
#include<stdio.h>
#include<math.h>
#include "nbody.h"
//#include "doub.h"
#include<stdlib.h>

//  Integrate calling kernels and dealing with all special cases for
// finding encounters

#define PPHERM 1  // if 1 put all particles in hermite integrator 
                  // when there is a planet planet encounter

#define CHECKENERGY 0


// forward declares
void copysublists(double* Pos1, double* Vel1, double* Pos0, double* Vel0,
   int* indexlist, int nencounter);
int lookplanetplanet(int *indexlist, int *startlist, int nlists, 
       int numPlanets);

#define KEPFAC 2   // reduces KEP KERNEL threads number

/// routine to do timestep
// aPos, aVel are new
// bPos, bVel are old
void
integrateKepCol(double* aPos, double* aVel,
                double* bPos, double* bVel,
                double deltaTime, double damping,  
                double soft2_int, double soft2_her,
                double hillfac_int, double hillfac_col,
                int numBodies, int numBodiesMassive, int numPlanets,
                int p, double m0, double time)
{


  double *Pos0, *Vel0;
  Pos0 = new double[numBodies*4];  // store positions +vels on CPU
  Vel0 = new double[numBodies*4];  // beginning of Kep/Hermite
  double *Pos1, *Vel1;
  Pos1 = new double[numBodies*4];  // store positions +vels on CPU
  Vel1 = new double[numBodies*4];  // ending of Kep/Hermite
  // these arrays also possibly used for coordinate transformation to and from
  // helio/bary coordinates 

  double es0;
  if(CHECKENERGY==1)
    // energy at beginning of timestep
    es0 = energysum(bPos,  bVel, soft2_int, numBodies, p, m0);

// starting kernel calls, drift step due to helio/bary coordinate system
  driftstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime/(2.0*m0), numBodies, p);

// interactions
  kickstepNbodySystem(bPos, bVel, aPos, aVel,
           deltaTime/2.0, damping,  soft2_int, 
           numBodies, numBodiesMassive, p,m0);
// incorrecly computed interactions will be removed later

// for bodies not in close approaches this below is the kepstep
// for bodies involved in close approaches these positions/vels 
//   will be used to find the close approaches
  kepstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime, numBodies, p/KEPFAC, m0);


//////////////////////////////
/// searching for close encounters
  int *indexlist,*startlist; // numBodies is the largest they can be
  indexlist = new int[numBodies]; // to store encounter lists
                                  // each list is consecutively stored
  startlist = new int[numBodies]; // startlist tell you where each encounter
                                  // list starts

//  int numPlanets=4; // used to specify how much to search for pairs encounters
                 // right now only searching for object/planet encounters
  int npairs;
  long long int* h_pairlist;
  int tnencounter=0;  /// total number in encounters

  // order Pos0,Vel0 then Pos1,Vel1
  // finding close approaches on gpu
  npairs = mkpairlist(bPos, bVel, aPos, aVel, numBodies, numBodiesMassive, 
       &h_pairlist, m0, m0, deltaTime, hillfac_col);
  //  printf("npairs found from mkpairlist = %d \n",npairs);

// we now have the newpairlist and can make indexlists from it
// this routine makes the indexlists for encounter sublists
  
  int nlists = nhermlists(indexlist, startlist, &tnencounter,
              numBodies, h_pairlist, npairs);

  free(h_pairlist);// remember to free 
                   // even if npairs==0, at least 1 element is allocated

  int planetplanet = 0;
  if (nlists>0){  // see if there are planet/planet encounters
    planetplanet = lookplanetplanet(indexlist, startlist, 
            nlists, numPlanets);
//    if (planetplanet ==1) printtocoll("PP");
//    if (planetplanet ==1) printf("planet planet encounter\n");
  }

// put all particles in hermite integrator if there is a planet/planet encounter
  if ((planetplanet==1) && (PPHERM ==1)) { //can turn this off with PPHERM
//    printallenclist(indexlist,startlist, nlists); AQ turned off!
    // print encounter list

    // go backwards to beginning of timestep
    kickstepNbodySystem(aPos, aVel, bPos, bVel,
           -deltaTime/2.0, damping,  soft2_int, 
            numBodies, numBodiesMassive, p,m0);
    
    driftstepNbodySystem(bPos, bVel, aPos, aVel,
           -deltaTime/(2.0*m0), numBodies, p);
    // we are now at beginning of timestep
    copyArrayFromDevice(Pos0, bPos, numBodies); // needed for hermite integrator
    copyArrayFromDevice(Vel0, bVel, numBodies);
    Pos0[3] = m0; // to compute energy, Hill rad etc, keep this
    double rc = 1e6; // make critical radius large!   
  
    // move to cartesian coordinates
    frombary(Pos1, Vel1, Pos0, Vel0, numBodies);
    int dmovesun=0; // allow sun to move, integrating in cartesian
    // call hermite integrator in cartesian coords for all particles
    hermiteBodies((double4*)Pos1, (double4*)Vel1,
       numBodies, deltaTime, soft2_her, rc, dmovesun,
       time,indexlist,numPlanets);
         // WRONG indexlist passed here
    // move back to bary/heliocentric coordinates
    tobary(Pos1, Vel1, Pos0, Vel0, numBodies);
    Pos0[3] = 0; 
    copyArrayToDevice(aPos, Pos0, numBodies);  // copy back the data
    copyArrayToDevice(aVel, Vel0, numBodies);  
    // we are done now, no further kernels to call but there is mem to free 
    // printf("end of planet/planet all on hermite\n"); fflush(stdout);
  }
  else{ // no planet/planet encounters or treating all encounters the same
    if (nlists>0){ // if there are encounters
      // no need to copy unless there are encounters
//      printf("encounters but not all on hermite\n");fflush(stdout);
      copyArrayFromDevice(Pos0, bPos, numBodies); // will be needed for hermite
      copyArrayFromDevice(Vel0, bVel, numBodies);
      copyArrayFromDevice(Pos1, aPos, numBodies); // will be needed for merge  
      copyArrayFromDevice(Vel1, aVel, numBodies); // for particles not 
                                                  // involved in encounters
      // Pos0 is beginning, Pos1 is end of Kepstep
      Pos0[3] = m0; // to compute energy, Hill rad etc, keep this
      Pos1[3] = m0; 
    
      // --- in sublist loop
      for(int i=0;i<nlists;i++){
         int offset = startlist[i];
         int nencounter =  startlist[i+1] - startlist[i];
//       printenclist(indexlist+offset, nencounter);fflush(stdout);
         // print out the encounter list

         // get maximum Hill sphere of encounter sublist
         double lh = callHill(indexlist+offset, nencounter,(double4*)Pos0,m0);
         double rc = hillfac_int*lh; // critical radius

         // remove interactions that should not have been computed in kickstep
         callIntCor(indexlist+offset, (double4*)Pos0,(double4*)Vel0, nencounter,
             deltaTime/2.0, soft2_int, rc); 

         // integrate close encounter list 
         int dmovesun=1; // don't move sun
         callHermite(indexlist+offset, (double4*)Pos0,(double4*)Vel0, 
              nencounter, deltaTime, soft2_her, rc, m0, dmovesun,time,
              numPlanets);

         // remove interactions that are about to be computed in kickstep
         callIntCor(indexlist+offset, (double4*)Pos0,(double4*)Vel0, 
            nencounter, deltaTime/2.0, soft2_int, rc); 

       
      } // --- exit sublist loop

      // merge hermite computed with the rest of the particles
      copysublists(Pos1, Vel1, Pos0, Vel0, indexlist, tnencounter);

      Pos0[3] = 0.0; Pos1[3] = 0.0;  // set central mass back to 0
      copyArrayToDevice(aPos, Pos1, numBodies);  // copy back the merged data
      copyArrayToDevice(aVel, Vel1, numBodies);  
   } // exit encounter loop

     // last few kernels to call
     // interactions if planetplanet!=1 but both for interactions and without
     kickstepNbodySystem(bPos, bVel, aPos, aVel,
           deltaTime/2.0, damping,  soft2_int, 
           numBodies, numBodiesMassive, p,m0);

     // drift step
     driftstepNbodySystem(aPos, aVel, bPos, bVel,
           deltaTime/(2.0*m0), numBodies, p);
   
  }

// cleanup
  delete [] indexlist;
  delete [] startlist;

  delete [] Pos0;
  delete [] Vel0;
  delete [] Pos1;
  delete [] Vel1;

  if(CHECKENERGY==1){
//  FILE *fpa = fopen("coll.txt","a");
    double es1 = energysum(aPos,  aVel, soft2_int, numBodies, p, m0);
    // energy at endof timestep
    double de = (es1-es0)/fabs(es0);
//  fprintf(fpa,"es1 = %.6e de=%.6e\n",es1,de); fflush(stdout);
//  fclose(fpa);
  }

}

// copy only the indexlist from Pos0,Vel0 to Pos1,Vel1
void copysublists(double* Pos1, double* Vel1, double* Pos0, double* Vel0,
   int* indexlist, int nencounter)
{
   for(int i=0;i<nencounter;i++){
      for(int j=0;j<4;j++){
         int index = indexlist[i]*4 +j;
         Pos1[index] = Pos0[index]; 
         Vel1[index] = Vel0[index]; 
      }
   }
}

// are there any planet/planet encounters present?
int lookplanetplanet(int *indexlist, int *startlist, int nlists, 
       int numPlanets)
{
  for(int i=0;i<nlists;i++){
     int offset = startlist[i];
     int nencounter =  startlist[i+1] - startlist[i];
     int onep =0;
     for(int j=0;j<nencounter;j++){
        int id = indexlist[j+offset];
        if (id<= numPlanets) onep++;
        if (onep>=2) return 1; 
     }
  }
  return 0;
}

