

//#include <paramgl.h>
#include <cstdlib>
#include <cstdio>
#include <algorithm>
#include <assert.h>
#include <math.h>
#include <string.h>
#include"nbody.h"

#define computeE 1 //if 1 compute energy on gpu at each output step

#define NTHREADS (int)128  // hardset here

// reads in parameter file param.in
// if you give it a number on the command line it does a restart at this timestep

// ouput filename manipulator
void mkofilename(char *pfile, char *ofile, int i);

int main(int argc, char *argv[])
{
   int threadsp; 
   threadsp = NTHREADS;
   // number of threads per block is p 
   // number of blocks is numBodies/p
   int numBodies;   // number of bodies in N-body simulation
   int numBodiesMassive;  // massive bodies in simulation
   // both numBodies and numBodiesMassive are integer multiples of threadsp
   SIMinfo siminfo; // structure for simulation information

//////////////////////////////////////////
// read in parameters
   // input parameter file name hardset right now
   sread("param.in",&siminfo);   // read in parameters to structure

// I want numBodies to be a multiple of threadsp
   int nmp = (int)ceil((float)siminfo.numBodies/threadsp);
   nmp *= threadsp;  // this should be a multiple of threadsp;
   numBodies = nmp;  // numBodies is a multiple of threadsp;            
   if (numBodies%threadsp != 0.0) {
      printf("please make numBodies (%d) a multiple of %d\n"
           ,numBodies,threadsp); // obsolete check
   }
   nmp = (int)ceil((float)siminfo.numBodiesMassive/threadsp);
   nmp *= threadsp;  // this should be a multiple of threadsp;
// I want numBodiesMassive to be a multiple of threadsp
   numBodiesMassive = nmp; // numBodiesMassive is now a multiple of threadsp;  
   if (numBodiesMassive > numBodies) numBodiesMassive = numBodies;
   
   printf("numBodies = %d from %d\n",numBodies,siminfo.numBodies);
   printf("numBodiesMassive = %d from %d\n"
        ,numBodiesMassive,siminfo.numBodiesMassive);
      
   FILE *fpo;
   if (argc==1) {  // if not a restart 
     fpo = fopen("coll.txt","w"); // empty this file if it exists
     fclose(fpo);
   }

   // CPU data
   double* hPos; // 4 vectors, x,y,z,m
   double* hVel; // 4 vector, vx,vy,vz,unused
   // GPU data
   double* m_dPos[2]; // recast as double4 during integration
   double* m_dVel[2]; // these are 2 times x,y,z,m and 2 times vx,vy,vz,unused
   // the two arrays are alternating current and past values for positions and velocities

   checkCUDA(siminfo.device); // device init

//  printf("soft2= %.2e\n",siminfo.soft2_int);
//  printf("damp= %.2f\n",siminfo.damping);
   printf("niter= %d\n",siminfo.niter);  // number of data outs
   printf("ndt= %d\n",siminfo.ndt);  // number of timesteps per dataout
   if (siminfo.soft2_int ==0.0){
      printf("warning: soft2_int ==0.0 and this may cause NaN problems \n");
   }

   // allocate CPU data
   hPos = new double[numBodies*4]; 
   hVel = new double[numBodies*4];
  // we allocate doubles  recast as double4 later on
   memset(hPos, 0, numBodies*4*sizeof(double));
   memset(hVel, 0, numBodies*4*sizeof(double));
   // hPos includes x,y,z,m 
   // hVel includes vx,vy,vz and last vector part unused 
   double* hPoshel; // 4 vectors, x,y,z,m  heliocentric coordinates
   double* hVelhel; // 4 vector, vx,vy,vz,unused barycentric coordinates
   hPoshel = new double[numBodies*4];    // new coordinates
   hVelhel = new double[numBodies*4];
   memset(hPoshel, 0, numBodies*4*sizeof(double));
   memset(hVelhel, 0, numBodies*4*sizeof(double));
   // because of memset particles with numBodies > siminfo.numBodies are zero

//////////////////////////////////
// load in initial particle positions and velocities
   int istart;
   char filename[300];
   double time;
   if (argc==1) {  // if run the first time at time =0
     istart = 0;
     strcpy(filename,siminfo.outroot);
     // read in the initial data from paramfile.h5part
#ifdef H5PART
     strcat(filename,".h5part");
     read_h5part(filename, hPos, hVel,  0); // get data
#else
     strcat(filename,".asc");
     read_asc(filename, hPos, hVel,  0); // get data
#endif 
     printf("data read in %s\n",filename); fflush(stdout);
     time=0.0; 
     // zero masses that are supposed to be zero
     for(int i = numBodiesMassive; i< numBodies; i++) hPos[4*i+3] = 0.0;
   }
   else { // restart on file istart
      sscanf(argv[1],"%d",&istart);
      mkofilename(siminfo.outroot, filename, istart); // create filename for iter step istart
#ifdef H5PART
      read_h5part(filename, hPos, hVel,  0); // get data, does not get time
#else
      read_asc(filename, hPos, hVel,  0); // get data, does not get time
#endif
      printf("data read in %s\n",filename); fflush(stdout);
      time=siminfo.ndt*siminfo.deltaTime*(double)istart; // time set here
// on restart parameters not read in from output data files but from param.in
   }
   siminfo.time = time;
   
   printf("We are about to allocate memory on GPU\n");
   // allocate double4 data on board GPU
   int err;
   err=allocateNBodyArrays(m_dPos, numBodies);
   err=allocateNBodyArrays(m_dVel, numBodies);
   if (err==1) exit(0); // quick exit if cudainit has not given an error

   // change  coordinates to heliocentric/barycentric
   double m0 = hPos[0+3]; // remember the central mass!
   correctcom(hPos, hVel, numBodies); // keep this for energy check
   tobary(hPos, hVel, hPoshel, hVelhel, numBodies); 
   // zero center of mass
   hPoshel[0+0]=0.0; hPoshel[0+1]=0.0; hPoshel[0+2]=0.0;
   // zero center of momentum
   hVelhel[0+0]=0.0; hVelhel[0+1]=0.0; hVelhel[0+2]=0.0;
   // zero central mass for integrations
   hPoshel[0+3]=0.0;

   printf("Mem allocated and coords trans\n");

// double cpusum = energy((double4*)hPos, (double4*)hVel, 
//       numBodies, siminfo.soft2_int,1e6); 
//   printf("cpusum=%.6e \n",cpusum); // agrees with gpusum!

   int iold=0;
   int inew=1; 
   // load initial particle positions and velocities onto device
   copyArrayToDevice(m_dPos[inew], hPoshel, numBodies); 
   copyArrayToDevice(m_dVel[inew], hVelhel, numBodies);
   copyArrayToDevice(m_dPos[iold], hPoshel, numBodies); // redundant ok
   copyArrayToDevice(m_dVel[iold], hVelhel, numBodies);

   // initial energy 
   
   double e0sum = energysum(m_dPos[inew],  
                      m_dVel[inew], siminfo.soft2_int, numBodiesMassive, 
	          	threadsp,m0); 
  

////////////////////////////////////////////////////////////
// iterations
   // only store output and compute energy each iteration
   int i=0;
   for (i=istart;i<siminfo.niter;i++){  // loop iterations
       // get data from device
       copyArrayFromDevice(hPoshel, m_dPos[inew], numBodies);
       copyArrayFromDevice(hVelhel, m_dVel[inew], numBodies);
       
       hPoshel[0+3]=m0; // restore central mass so coordinate conversion done ok
       frombary(hPos, hVel, hPoshel, hVelhel, numBodies); // back to cartesian coords
      
       mkofilename(siminfo.outroot, filename, i); // create output filename
       if ((i>istart)||(i==0)) // don't write over if restart
#ifdef H5PART
         write_h5part(filename, hPos, hVel, siminfo, 0);  // write h5part file
#else
         write_asc(filename, hPos, hVel, siminfo, 0);  // write ascii file
#endif

       // compute energy from gpu calculated energy vector
       if (computeE) {
//       cpusum = energy((double4*)hPos, (double4*)hVel, 
//                    numBodies, siminfo.soft2_int,1e6); 
         double esum = energysum(m_dPos[inew],  
                      m_dVel[inew], siminfo.soft2_int, numBodiesMassive, 
	          	threadsp,m0); 
         double de =  (esum-e0sum)/fabs(e0sum);
         printf("%d ensum=%.8e de/e=%.5e P=%.4f \n"
            ,i, esum, de,time/(2.0*M_PI)); fflush(stdout);
         if (isnan(esum)) exit(0);
       }

        ////////////////////////////////////////////
        // timesteps in each iteration
       for (int j=0;j<siminfo.ndt;j++){  // loop timesteps for each iteration
         std::swap(iold,inew); //  set new to old
         // do integration
/* // integrate without encounters
         integrateKepNbodySystem(
               m_dPos[inew], m_dVel[inew], m_dPos[iold], m_dVel[iold],
                     siminfo.deltaTime, siminfo.damping,  siminfo.soft2_int,
                     numBodies, numBodiesMassive, threadsp, m0);
*/
  // integrate with encounters
         integrateKepCol(
                m_dPos[inew], m_dVel[inew], m_dPos[iold], m_dVel[iold],
                siminfo.deltaTime, siminfo.damping,
                siminfo.soft2_int, siminfo.soft2_her,
                siminfo.hillfac_int, siminfo.hillfac_col,
                numBodies, numBodiesMassive, siminfo.nplanets, 
                threadsp, m0,time);

         
         time += siminfo.deltaTime;
         fpo = fopen("coll.txt","a"); // close approach list file
         fprintf(fpo,"period= %.4f\n",time/(2.0*M_PI));
         fclose(fpo);
      }
      siminfo.time = time;
   }
// store last timestep
       copyArrayFromDevice(hPoshel, m_dPos[inew], numBodies);
       copyArrayFromDevice(hVelhel, m_dVel[inew], numBodies);
       hPoshel[0+3]=m0;
       frombary(hPos, hVel, hPoshel, hVelhel, numBodies); // back to cartesian coords
       mkofilename(siminfo.outroot, filename, i); // create output filename
#ifdef H5PART
       write_h5part(filename, hPos, hVel, siminfo, 0);  // write h5part file
#else
       write_asc(filename, hPos, hVel, siminfo, 0);  // write ascii  file
#endif
       if (computeE) {
//       cpusum = energy((double4*)hPos, (double4*)hVel, 
//                    numBodies, siminfo.soft2_int,1e6); 
         double esum = energysum(m_dPos[inew], 
                      m_dVel[inew], siminfo.soft2_int, numBodiesMassive, 
		threadsp, m0); 
         double de =  (esum-e0sum)/fabs(e0sum);
         printf("%d ensum=%.8e de/e=%.5e P=%.4f \n"
            ,i, esum, de,time/(2.0*M_PI)); fflush(stdout);
       }

   // cleanup
   deleteNBodyArrays(m_dPos);
   deleteNBodyArrays(m_dVel);
   delete [] hPos;
   delete [] hVel;
   delete [] hPoshel;
   delete [] hVelhel;

}
// create filenames for out data files
void mkofilename(char *pfile, char *ofile, int i)
{
  char junks[20];
  strcpy(ofile,pfile);
  strcpy(junks,"_");
  if (i<1000) strcat(junks,"0");
  if (i<100)  strcat(junks,"0");
  if (i<10)   strcat(junks,"0");
  strcat(ofile,junks);
  sprintf(junks,"%d",i);
  strcat(ofile,junks);
#ifdef H5PART
  strcat(ofile,".h5part");
#else
  strcat(ofile,".asc");
#endif 
//  printf("mko %s\n",ofile); fflush(stdout);
}

 
